Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely hide atomic autosplit finfun.
Require Import pperm privstate footprint model lemmas stsep histories.
Require Import treiber stack_framed getter alloc Eqdep state.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.


Section TreiberSequential.
Variables (pv al : nat) (T : treiberT) (asent alk : ptr).
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation tb := (tb_label T).
Notation st := [encoded_set of seq A].
Notation U := (tbU T). 
Notation tb_getso s := (getso (tbg T) s).
Notation tb_self s := (self (tbg T) s).
Notation tb_other s := (other (tbg T) s).
Notation tb_joint s := (joint (tbg T) s). 
Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s).
Notation pv_other s := (other (pvg pv) s).
Notation pv_coh := (priv_coh pv).
Notation tb_coh := (TreiberCoh.Exports.tb_coh T).
Notation l := (tb_label T).

(* Use a concrete, CAS-based allocator *)
Notation AllocM := (AllocCAS al asent alk). 
Notation V := (entangle (Priv pv) AllocM). 
Notation al_coh := (al_coh al asent alk). 

(************************ Auxiliary lemmas  *************************)
Definition cast T (t t' : mod) (r : t = t') (e : T t) :=
  match r in (_ = t') return T t' with erefl => e end.

Lemma eqc S t (r : t = t) (e : S t) : cast r e = e.
Proof. by move: r; apply: Streicher_K. Qed.

Lemma val_cast (V W: mod) (A: Type) (pf: V = W) i (e: ST V A) (r: model.cont A) :
  verify (W:=V) i e r -> 
  verify (W:=W) i (cast (T:=fun m => ST m _) pf e) r. 
Proof. by move: (pf); rewrite pf in e *=>{pf}pf; rewrite eqc. Qed.

Lemma validSt sp sa (sh: heap) (lp ep : U) :
  valid (sp \+ sa) -> 
  l \notin sdom (sp \+ sa) ->
  valid (sp \+ sa \+ l ->> [lp, sh, ep] : statePCM).
Proof.
move=>V Nl.
rewrite /valid !fsE /= -!(joinC (l \->_)) -(joinC (l ::->_)) !fvalidPtUn.
move: V; rewrite {1}/valid /=; case/and3P=>->->-> /=.
by apply/and3P; split; apply:contra Nl=>H; rewrite inE; apply/or3P;
   [constructor 1 | constructor 2 | constructor 3].
Qed.

Lemma validPvSubst (hL hL' hE : heap) sa:
  valid (pv ->> [hL, tt, hE] \+ sa) ->
  valid (pv ->> [hL', tt, hE] \+ sa).
Proof.
rewrite /valid !fsE /=; case/and3P=>H1 H2 H3.
rewrite !fvalidPtUn in H1 H2 H3 *.
by apply/and3P; split=>//.
Qed.

Lemma notinPvSubst lb (hL hL' hE : heap) sa:
  lb \notin sdom (pv ->> [hL, tt, hE] \+ sa) ->
  lb \notin sdom (pv ->> [hL', tt, hE] \+ sa).
Proof.
move=>Nd; apply: contra Nd. 
rewrite !fsE !inE /=; case/or3P=>H; apply/or3P; first last;
[by constructor 3|by constructor 2|constructor 1];
move: H; rewrite !fdomUn !inE /= !inE /=; case/andP=>H1->.
by apply/andP; split=>//; rewrite !fvalidPtUn in H1 *.
Qed.

(********************** End of auxiliary lemmas  ***********************)

Definition push_seq_tp (e: A) :=
  {ls}, STsep [V] (fun i => exists (p: ptr) hs, 
                      [/\ pv_self i = sent :-> p \+ hs,
                          l \notin sdom i & 
                          llist p ls hs],
                     fun (_ : unit) m => exists (p: ptr) hs g, 
                       [/\ pv_self m = sent :-> p \+ hs \+ g & 
                           llist p (e::ls) hs]). 

Program Definition push_seq e : push_seq_tp e :=
  Do (priv_hide (TreiberDecorate T) (Unit, Unit) 
                (cast (T:=fun m => ST m _) (entangleAC _ _ _) 
                      (push_framed pv al asent alk e))).
Next Obligation.
apply: gh=>i1 ls[p][hs][Ep1] Nd1 L1 C1. 
case: (C1)=>_[sp1][sa1][Zi1] Cp1 Ca1; subst i1; rewrite (getC Cp1 C1) /= in Ep1.
case: (Cp1)=>_[hL][hE] Zp1; rewrite {sp1}Zp1 getE in Ep1 Cp1 C1 Nd1 *.  
set st1 := l ->> [0 \\-> (ls, ls) : U, sent :-> p \+ hs, Unit].
have X : flats st1 = sent :-> p \+ hs by rewrite tb_flatsE.
have V : valid (pv ->> [hL, tt, hE] \+ sa1 \+ st1) by apply: validSt (cohV C1) Nd1. 
have V' : valid (pv ->> [Unit, tt, hE] \+ sa1 \+ st1). 
- by rewrite /valid !fsE /= -!joinA !fvalidPtUn in V *.
rewrite -(unitL hL) Ep1 -X; apply: val_privhide'=>//.
- exists hs, ls, ls, p; rewrite !unitR; split=>//.
  exists p, hs, Unit, ls; rewrite !unitR !getE !unitR; split=>//.
  - by rewrite last_keyPt um_findPt; exists ls. 
  - by split; [apply: contPt | apply: complete0].
  move: (cohH Cp1); rewrite /validH tb_flatsE priv_flatsE -Ep1.
  by move/validL.
apply: val_cast; apply: (gh_ex Unit); apply: (gh_ex (0 \\-> (ls, ls): U)).
apply: (gh_ex Unit); apply: val_do=>[_|_ m [t][ls']].
- by rewrite !getV //= unitR; split=>//; apply: hist_unit; rewrite um_validPt.
case=>Et Ep _ _ Cm; case: (Cm)=>_ [spt2][sa2][Em] Cm2 Ca2; subst m. 
case: Cm2=>_ [sp2][st2][E2]Cp2 Ct2; subst spt2=>i' j' hL' hE' g' Em Td.
rewrite Em in Cm Et Ep=>{Em Ct2 st2 Cp2 sp2 Ca2 sa2}.
have V2 : valid (pv ->> [hL', tt, hE'] \+ i' \+ j') by apply: cohV Cm.
have V3 : valid (pv ->> [hL' \+ flats j', tt, hE'] \+ i'). 
- by move/validL: V2; rewrite /valid !fsE /= !fvalidPtUn. 
case: Td=>h [ls0][ls1][p'][[Ej]L Cj]; rewrite {j'}Ej  in Cm Cj Et Ep V2 V3 *;
rewrite !getV //= in Ep Et *; rewrite {V2 V3}tb_flatsE.
have Y : valid (0 \\-> (ls0, ls0) \+ g'.1).
- by move/cohSO: Cj; rewrite /validSO fsE /= zipPt unitR allvalidPt.
case/(um_cancel Y): Et; case=>? _ W G; subst ls0.
exists p', h, g'.2; rewrite Ep unitL; split=>//; rewrite {}G in Y W Cj Cm.
case: (TreiberDecorate.hist_init1 _ Cj); first by rewrite um_empbPt.
move=>x /um_findPt_inv [?[??]]; subst t ls' x.
have K : last_key (0 \\-> (ls, ls) \+ 1 \\-> (ls, e::ls)) = 1.
- rewrite (_ : 1 = fresh (0 \\-> (ls, ls))) ?last_fresh // ?um_validPt //.
  by rewrite /fresh last_keyPt.
suff : llist p' (e :: ls) h by [].
move: (cohV Cj)=>Vj; case: (Cj); rewrite !getV //=.
move=>p2 [h2][g2][ls2][_] E2 L2 X2 [_] _. 
rewrite /validH tb_flatsE => V2; rewrite -!joinA in E2 V2.
case/(hcancelV V2): E2=>? V3 E3. subst p2.
case: (llist_prec V3 E3 L L2)=>??? {E3 Vj}; subst ls2 h2 g2.
case: X2=>x; rewrite unitR K; rewrite joinC in Y *; rewrite um_findPtUn //.
by case=>??; subst ls1 x.
Qed.

Definition pop_seq_tp :=
  {ls}, STsep [entangle (Priv pv) empty_mod] 
  (fun i => exists (p: ptr) hs, 
    [/\ pv_self i = sent :-> p \+ hs, l \notin sdom i & llist p ls hs],
   fun (t : option A) m => if t is Some e 
     then exists (p: ptr) ls' hs g, 
          [/\ pv_self m = sent :-> p \+ hs \+ g, ls = e::ls' & llist p ls' hs]
     else exists g, [/\ pv_self m = sent :-> null \+ g & ls = [::]]). 

Program Definition pop_seq : pop_seq_tp :=
  Do (priv_hide (TreiberDecorate T) (Unit, Unit) (extend _ (pop_framed T))).
Next Obligation. exact: injRE (erefl _). Defined. 

Next Obligation.
apply: gh=>i1 ls[p][hs][Ep1] Nd1 L1 C1. 
case: (C1)=>_[sp1][sa1][Zi1] Cp1 Ca1; subst i1; rewrite (getC Cp1 C1) /= in Ep1.
case: (Cp1)=>_[hL][hE] Zp1; rewrite {sp1}Zp1 getE in Ep1 Cp1 C1 Nd1 *.  
set st1 := l ->> [0 \\-> (ls, ls) : U, sent :-> p \+ hs, Unit].
have X : flats st1 = sent :-> p \+ hs by rewrite tb_flatsE.
have V : valid (pv ->> [hL, tt, hE] \+ sa1 \+ st1) by apply: validSt (cohV C1) Nd1. 
have V' : valid (pv ->> [Unit, tt, hE] \+ sa1 \+ st1). 
- by rewrite /valid !fsE /= -!joinA !fvalidPtUn in V *.
have C: st1 \In tb_coh.
- exists p, hs, Unit, ls; rewrite !unitR !getE !unitR; split=>//.
  - by rewrite last_keyPt um_findPt; exists ls. 
  - by split; [apply: contPt | apply: complete0].
  move: (cohH Cp1); rewrite /validH tb_flatsE priv_flatsE -Ep1.
  by move/validL.
rewrite -(unitL hL) Ep1 -X; apply: val_privhide'=>//.
- exists hs, ls, ls, p; rewrite !unitR; split=>//.
rewrite joinC; apply: val_extend=>//; rewrite /st1 in C *.
apply: (gh_ex Unit); apply: (gh_ex (0 \\-> (ls, ls))). 
apply: val_do=>[_|v m [_]].
- rewrite !getV //= ?um_validPt ?unitR; do?[apply: (cohV C)].
  by split=>//; apply: hist_unit; rewrite um_validPt.
move=>XX _ j' /= Cm' M i j2 hL' hE' [hst hs2] Em; rewrite starC0 in Cm'.
case=>hs' [ls0][ls2][p'] /= [[Ej2 L2]] Cj2; rewrite Em in Cm'.
move/(star_coh_prec Cm'): (Cj2)=>Cp2; rewrite -Em in Cm'.
case: (menvs_coh M)=>_ /= Cj2'; rewrite starC0 in Cj2'.
rewrite joinC in Em Cm'.
move: (coh_prec (cohV Cm') Em Cj2' Cp2)=>E; rewrite {j'}E in Cj2' M Em Cm'.
move/(joinsK (cohV Cm')): Em=>Em; subst j2.
rewrite tb_flatsE; rewrite {m}Em in Cm' XX *; rewrite !getE in XX.
rewrite !getE; last first.
- by move/validL: (cohV Cm'); rewrite /valid /= !fsE /= !fvalidPtUn. 
case: v XX=>[v[t][ls'][_]|]; last first.

(* Nil case *)
- rewrite unitR; case=>t [_] E; rewrite {ls0 X V V'}E in Cm' Cj2.
  case=>x' /um_findPt_inv [Z1]; case=>Z2 Z3; subst t x' ls.
  case: Cj2; rewrite !getE !unitR=>p2[h][g][ls][_]E L H1 _.
  rewrite /validH tb_flatsE=>Vh.
  case:H1=>x' /um_findPt_inv [_]; case=>Z2 Z3; subst x' ls.
  case:L=>Z1 Z2; subst p2 h; rewrite unitR -joinA in E Vh.
  case/(hcancelV Vh): E=>Z1 _ Z2; subst p' g; rewrite joinC -!joinA.
  by exists (hs' \+ (hs2 \+ hL')).

(* Cons case *)
move/cohSO: (Cj2); rewrite /validSO fsE /= zipPt unitR allvalidPt=>V1.
case/(um_cancel V1); case=>Z1 _ _ Z2; subst ls0 hst.
case: (Cj2); rewrite !getE !unitR=>p2[h][g][ls1][_]E L H1 [H2 H3].
rewrite /validH tb_flatsE=>Vh.
have H4: last_key (0 \\-> (ls, ls) \+ t \\-> (v :: ls', ls')) = t.
- apply: max_key_last=>[|y]; last first.
  - by rewrite domUn -topredE V1/=!um_domPt -!topredE/==>/orP; case=>/eqP<-. 
  by rewrite domUn -topredE V1 /=; apply/orP; right; rewrite um_domPt -topredE /=. 
rewrite H4 joinC in H1 V1=>{H4}.
case: H1=>ls0; rewrite (um_findPtUn V1); case=>Z1 Z2; subst ls0 ls1.
case: (TreiberDecorate.hist_init1 _ Cj2)=>[|ls1]; first by rewrite um_empbPt.
move/um_findPt_inv=> [?[??]]{X V V'}; subst t ls ls1.
rewrite -!joinA in E Vh *; case/(hcancelV Vh): E=>Z1 _ Z2; subst p2; rewrite Z2.
by rewrite joinC -!joinA joinA; exists p', ls', h, (g \+ hL').
Qed.

End TreiberSequential.



