Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely hide atomic autosplit.
Require Import privstate footprint model lemmas stsep extra_lemmas histories.
Require Import getter alloc.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Notation "[ e ]" := (extend _ e).

(************************************)
(* first, stuff about linked lists  *)
(* this should be a generic library *)
(************************************)

Fixpoint llist A p (xs : seq A) {struct xs} := 
  if xs is x::xt then 
    [Pred h | exists r h', h = p :-> (x, r) \+ h' /\ h' \In llist r xt]
  else [Pred h | p = null /\ h = Unit].

Lemma llist_prec A p (l1 l2 : seq A) h1 h2 g1 g2 :
        valid (h1 \+ g1) ->
        h1 \+ g1 = h2 \+ g2 -> 
        llist p l1 h1 -> llist p l2 h2 -> 
        [/\ l1 = l2, h1 = h2 & g1 = g2].
Proof.
move=>V E H1 H2; elim: l1 l2 h1 h2 p H1 H2 E V=>[|a1 l1 IH].
- case=>[|a2 l2] h1 h2 p /= [->->].
  - by case=>_ ->; rewrite !unitL. 
  by case=>r [h'][-> _ ->] /validL /hvalidPt_cond.
case=>[|a2 l2] h1 h2 p /= [p1][k1][-> H1].
- by case=>->-> _ /validL /hvalidPt_cond. 
case=>p2 [k2][-> H2]; rewrite -!joinA => E V.
case: {E V} (hcancelV V E) H1 H2; case=>->-> V E H1 H2. 
by case: (IH _ _ _ _ H1 H2 E V)=>->->->. 
Qed.

Lemma first_exists A p h (ls : seq A) : 
        p != null -> llist p ls h ->
        exists x r h', [/\ ls = x :: behead ls, h = p :-> (x, r) \+ h' & llist r (behead ls) h'].
Proof.
case: ls=>[|a ls] /= H []; first by case: eqP H.
by move=>r [h'][-> H1]; eexists a, r, h'.
Qed.

Lemma llist_null A (xs : seq A) h : 
        valid h -> llist null xs h -> xs = [::] /\ h = Unit.
Proof.
case: xs=>[|x xs] /= V H; first by case: H. by case: H V=>p [h'][-> _] /hvalidPt_cond.
Qed. 


(*************************)
(* Now the Treiber stuff *)
(*************************)


(* A concurroid for the Treiber stack *)
Structure treiberT := TreiberT {
  tb_label : nat;
  tb_set : encoded_set;   (* type of elements *)
  tb_sent : ptr           (* sentinel pointer *)
}.

Definition tbU T := [encoded_pcm of hist [encoded_set of seq (tb_set T)]].

Definition P_treiber T s1 s2 := exists e : tb_set T, s1 = e::s2 \/ s2 = e::s1.

Definition stacks_cont T := cont (@P_treiber T).

Module TreiberCoh.
Section TreiberCoh.
Variable T : treiberT.
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation tb := (tb_label T).
Notation st := [encoded_set of seq A].
Notation U := (tbU T). 

Definition tbg := @Getter tb U heap_tp_Encoded Unit.

Notation getso s := (getso tbg s).
Notation self s := (self tbg s).
Notation other s := (other tbg s).
Notation joint s := (joint tbg s).

Definition coh := 
  [Pred s | let: H := self s \+ other s in
            exists p h g ls,
              [/\ s = tb ->> [self s, joint s, other s], 
                  joint s = sent :-> p \+ h \+ g,
                  llist p ls h, [ls in H at last_key H],
                  stacks_cont H /\ complete H & 
                  validH s]].

Lemma tb_flatsE (hs ho : hist st) (h : heap) : flats (tb ->> [hs, h, ho]) = h.
Proof. by rewrite fsE flatsE !pcm_flatPt tp_flatPtE unitL unitR. Qed.

Lemma tb_cohE s : s \In coh -> s = tb ->> [self s, joint s, other s].
Proof. by case=>p [h][g][ls][]. Qed.

(* the coherence predicate lemmas *)

Lemma tb_cohSO s : s \In coh -> validSO s. 
Proof. 
case=>p [h][g][ls][E _ _ _ [_][[H _]] _]; rewrite /validSO E !fsE zipPt allvalidPt.
by case: H=>s' /find_some /dom_valid. 
Qed.

Lemma tb_cohH s : s \In coh -> validH s.
Proof. by case=>p [h][g][ls][]. Qed.

Lemma tb_cohfV s : s \In coh -> footV_eq s (mem_seq' [:: tb]).
Proof. by move/tb_cohE=>->; rewrite fsE. Qed.

Lemma tb_coh_weq s s' : s \In coh -> s' \In coh -> weq typeCode (shr s) (shr s').
Proof. by move/tb_cohE=>-> /tb_cohE ->; rewrite !fsE /=. Qed.

Lemma tb_coh_sig s : s \In coh -> tb \-> (Unit : hist st) = [zunit loc s].
Proof. by move/tb_cohE=>->; rewrite fsE zunitPt. Qed.

Lemma tb_coh_zigzag s p : s \< p \In coh <-> s \> p \In coh.
Proof.
split=>C; case: C (C)=>q [h][g][ls][H1 H2 H3 H4 [H5 H6] Vh] C;
have Vso : validSO (s \< p)
 by [rewrite ?(tb_cohSO C) // ?validSO_zigzag ?(tb_cohSO C)].
- rewrite validH_zigzag // in Vh; rewrite zagzigVE // in H4 H5 H6.
  move: H1; rewrite zagzigE => H1. 
  by exists q, h, g, ls; split=>//; rewrite !getE in H1 H2 *. 
rewrite -validH_zigzag // in Vh; rewrite -zagzigVE // in H4 H5 H6. 
move: H1; rewrite -zagzigE /= => H1. 
by exists q, h, g, ls; split=>//; rewrite !getE in H1 H2 *. 
Qed.

Definition TreiberCohMixin :=
  CohPredMixin tb_cohSO tb_cohH tb_cohfV tb_coh_weq tb_coh_sig tb_coh_zigzag. 
Definition tb_coh := Eval hnf in CohPred coh TreiberCohMixin.

Definition TreiberGCMixin := @GCMixin _ tb_coh tb_cohE.
Canonical tb_getcoh := Eval hnf in GC tbg tb_coh TreiberGCMixin.

End TreiberCoh.

Module Exports.
Notation tb_coh := tb_coh.
Notation tb_flatsE := tb_flatsE.
Notation tbg := tbg.
Notation tb_getcoh := tb_getcoh.
Canonical tb_getcoh.

(* additional lemma about Treiber joint *)

Lemma tb_cohJ T s : s \In tb_coh T -> valid (joint (tbg T) s).
Proof. by move=>C; move: (cohH C); rewrite {1}(cohE C) /validH tb_flatsE. Qed.

End Exports.
End TreiberCoh.

Export TreiberCoh.Exports.


Module PopTrans.
Section PopTrans.
Variable T : treiberT.
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation coh := (tb_coh T).
Notation U := (tbU T).
Notation getso s := (getso (tbg T) s).
Notation self s := (self (tbg T) s).
Notation other s := (other (tbg T) s).
Notation joint s := (joint (tbg T) s).

(* pop-relation doesn't release the heap *)
Definition pop_rel s s' :=
  let: H := self s \+ other s in
  exists (p p' : ptr) e (h : heap) ls, 
    [/\ joint s = sent :-> p \+ p :-> (e, p') \+ h,
        joint s' = sent :-> p' \+ p :-> (e, p') \+ h, 
        self s' = self s \+ fresh H \\-> (e :: ls, ls),
        other s' = other s,
        s \In coh & s' \In coh].

Lemma pop_coh s s' : pop_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>p [p'][e][h][ls][]. Qed.

Lemma pop_env s s' : pop_rel s s' -> env s = env s'.
Proof. 
case=>p [p'][e][h][ls][_ _ _ E C C']. 
by rewrite (cohE C) (cohE C') !fsE /= E.
Qed.

Lemma pop_zagzig s s' p : 
        env s = env s' ->
        pop_rel (s \> p) (s' \> p) -> pop_rel (s \< p) (s' \< p).
Proof.
move=>E [q][p'][e][h][ls][J1 J2 Es _ /coh_zigzag C /coh_zigzag C'].
exists q, p', e, h, ls; rewrite !getE !getzC // in J1 J2 Es *.
by split=>//; [rewrite joinAC Es joinA | rewrite E].
Qed.

Lemma pop_footH s s' : pop_rel s s' -> footeqH (flats s) (flats s').
Proof.
case=>q [p'][e][h][ls][J1 J2 _ _ C C'].
rewrite (cohE C) (cohE C') !tb_flatsE {}J1 {}J2 -!joinA. 
by apply: footeqHUnL; apply: footeqHPt.
Qed.

End PopTrans.

Module Exports.

Definition PopGuarMixin tbT := 
  GuarMixin (@pop_coh tbT) (@pop_env tbT) (@pop_zagzig tbT). 
Canonical pop_guar tbT := 
  Eval hnf in Guar (pop_rel tbT) (PopGuarMixin tbT).

Definition PopIntGuarMixin tbT := IntGuarMixin (@pop_footH tbT).
Canonical PopIntGuar tbT := 
  Eval hnf in IntGuar (pop_guar tbT) (PopIntGuarMixin tbT).

End Exports.
End PopTrans.

Export PopTrans.Exports.


Module PushTrans.
Section PushTrans.
Variable T : treiberT.
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation coh := (tb_coh T).
Notation U := (tbU T).
Notation getso s := (getso (tbg T) s).
Notation self s := (self (tbg T) s).
Notation other s := (other (tbg T) s).
Notation joint s := (joint (tbg T) s).

Variable h' : heap.

(* push_rel implements an acquire transition *)
Definition push_rel s s' :=
  let: H := self s \+ other s in
  exists (p p' : ptr) e (h : heap) ls, 
    [/\ h' = p' :-> (e, p),
        joint s  = sent :-> p \+ h,
        joint s' = sent :-> p' \+ p' :-> (e, p) \+ h, 
        self s' = self s \+ fresh H \\-> (ls, e::ls),
        other s' = other s &  
        s \In coh /\ s' \In coh].

Lemma push_coh s s' : push_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>p [p'][e][h][ls][]. Qed.

Lemma push_env s s' : push_rel s s' -> env s = env s'.
Proof. 
case=>? [?][?][?][?][_ _ _ _ E [C C']]. by rewrite (cohE C) (cohE C') !fsE /= E.
Qed.

Lemma push_zagzig s s' p : 
        env s = env s' -> push_rel (s \> p) (s' \> p) -> push_rel (s \< p) (s' \< p).
Proof.
move=>E [q][p'][e][h][ls][H J1 J2 Es _ [/coh_zigzag C /coh_zigzag C']].
exists q, p', e, h, ls; rewrite !getE !getzC // in J1 J2 Es *.
by split=>//; [rewrite joinAC Es joinA | rewrite E].
Qed.

Lemma push_footH s s' : push_rel s s' -> footeqH (flats s \+ h') (flats s').
Proof.
case=>q [p'][e][h][ls][H J1 J2 _ _ [C C']].
rewrite (cohE C) (cohE C') !tb_flatsE {}J1 {}J2 H.
by rewrite -!joinA -(joinC h); apply: footeqHUnL; apply: footeqHPt.
Qed.

(* the dual release transition doesn't release anything *)
(* i.e., the transfered heap is always empty *)
(* so, it's a dummy transition *)
Definition dummy_rel s s' := [/\ h' = Unit, s = s' & s \In coh].

Lemma dummy_coh s s' : dummy_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>_ ->. Qed.

Lemma dummy_env s s' : dummy_rel s s' -> env s = env s'.
Proof. by case=>_ ->. Qed.

Lemma dummy_zagzig s s' p : 
        env s = env s' ->
        dummy_rel (s \> p) (s' \> p) -> dummy_rel (s \< p) (s' \< p).
Proof. by move=>E [H /(zagzig_eq E) ->] /coh_zigzag. Qed.

Lemma dummy_footH s s' : dummy_rel s s' -> footeqH (flats s) (h' \+ flats s').
Proof. by case=>->->; rewrite unitL. Qed.

End PushTrans.

Module Exports.

Definition PushGuarMixin T h := 
  GuarMixin (@push_coh T h) (@push_env T h) (@push_zagzig T h). 
Canonical push_guar T h := 
  Eval hnf in Guar (push_rel T h) (PushGuarMixin T h).

Definition PushAcqGuarMixin T := AcqGuarMixin (@push_footH T).
Canonical PushAcqGuar T := Eval hnf in AcqGuar (push_guar T) (PushAcqGuarMixin T).

Definition DummyGuarMixin T h := GuarMixin (@dummy_coh T h) (@dummy_env T h) (@dummy_zagzig T h).
Canonical dummy_guar T h := Eval hnf in Guar (dummy_rel T h) (DummyGuarMixin T h).

Definition DummyRelGuarMixin T := RelGuarMixin (@dummy_footH T).
Canonical DummyRelGuar T := Eval hnf in RelGuar (dummy_guar T) (DummyRelGuarMixin T).

End Exports. End PushTrans.

Export PushTrans.Exports.

(* building the concurroid *)

Definition TreiberChanCore T := ChanCore (PushAcqGuar T) (DummyRelGuar T).
Canonical TreiberCore T := @Mod.Core (tb_coh T) (Pred1 (PopIntGuar T)).
Canonical Treiber T := @Mod.Make (TreiberCore T) (Pred1 (TreiberChanCore T)). 


(***********)
(* Actions *)
(***********)

Module TryPopAct.
Section TryPopAct.
Variables (T : treiberT) (p p' : ptr).
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation tb := (tb_label T).
Notation coh := (tb_coh T).
Notation U := (tbU T).
Notation getso s := (getso (tbg T) s).
Notation self s := (self (tbg T) s).
Notation other s := (other (tbg T) s).
Notation joint s := (joint (tbg T) s).
Notation guar := (Mod.internal (Treiber T)). 

Definition trypop_safe s := 
  exists (q : ptr) (e : A) h, 
    [/\ p != null, joint s = sent :-> q \+ p :-> (e, p') \+ h & s \In coh].

Definition trypop_step s s' v :=
  let: H := self s \+ other s in
  exists (q : ptr) e h, 
    [/\ p != null, joint s = sent :-> q \+ p :-> (e, p') \+ h,
        other s' = other s, v = (p == q),
        if v then 
          exists ls, joint s' = sent :-> p' \+ p :-> (e, p') \+ h /\
                     self s' = self s \+ fresh H \\-> (e::ls, ls)
        else s' = s & s \In coh /\ s' \In coh].

Lemma trypop_safe_coh : safe_coh coh trypop_safe.
Proof. by move=>s [q][e][h][]. Qed.

Lemma trypop_step_safe : step_safe trypop_safe trypop_step.
Proof. by move=>s s' v [q][e][h][????? [C _]]; exists q, e, h. Qed.

Lemma trypop_safe_zagzig : safe_zagzig trypop_safe.
Proof. 
move=>s t [q][e][h][Hp J /coh_zigzag C]. 
by exists q, e, h; rewrite !getE in J *.
Qed.

Lemma trypop_step_guar : step_guar guar trypop_step.
Proof. 
move=>s s' v [q][e][h][Hp J Eo Ev H [C C']].
case: (p =P q) Ev H J => [<-|_] ->; last by left.
case=>ls [H]; right; eexists _; do !split=>//. by exists p, p', e, h, ls. 
Qed.

Lemma trypop_step_total : step_total trypop_safe trypop_step.
Proof. 
move=>s [q][e][h'][Hp E C].
case X : (p == q); last by exists s, false, q, e, h'. 
rewrite -{q X}(eqP X) in E *.
set H : hist _ := self s \+ other s.
move: (tb_cohJ C) (cohVSO C)=>/= V Vso.
(* existential hell; can be ameliorated with getters for sentinel and heap *)
have [ls [h [g [Es Eh Hk Hs Hc Hl]]]] : exists ls h g,
  [/\ s = tb ->> [self s, joint s, other s], h' = h \+ g, 
      [e :: ls in H at last_key H], stacks_cont H, complete H & llist p' ls h].
- case: C=>q [k][g][ls][Es] Ej Hl Hk [Hs Hc] _.
  rewrite E -!joinA in V Ej; case/(hcancelV V): Ej Hl=><-{q} {V} V Ej Hl.
  case/(first_exists Hp): Hl Ej=>e' [r][h][E' ->{k} Hl Ej].
  rewrite -!joinA in Ej; case/(hcancelV V): Ej E' Hl; case=><-<- _ X E'.
  by exists (behead ls), h, g; rewrite -E'.
rewrite {h'}Eh in E.
set hs' : hist _ := self s \+ fresh H \\-> (e :: ls, ls).
set hj' : heap := sent :-> p' \+ p :-> (e, p') \+ (h \+ g).
set ho' := other s; set s' := tb ->> [hs', hj', ho'].
suff C' : s' \In coh.
- exists s', true, p, e, (h \+ g).
  by rewrite !getE // eq_refl; split=>//; exists ls.
exists p', h, (p :-> (e, p') \+ g), ls; rewrite /s' !getE; split=>//.
- by rewrite -joinA (joinCA h) joinA. 
- exists (e :: ls); rewrite joinAC last_fresh //.
  by rewrite findUnR ?valid_fresh // um_domPt inE eq_refl um_findPt. 
- split; first by rewrite joinAC; apply: fresh_cont=>//; exists e; left.
  by rewrite joinAC; apply: fresh_complete.
by rewrite /validH tb_flatsE /hj' E -!joinA !hvalidPtUn in V *.
Qed.

Lemma trypop_step_frame : step_frame trypop_safe trypop_step.
Proof. 
move=>s s' t v _ [q][e][h][Hp Ej Eo -> H [C C']]; 
rewrite !getE getzC //= in Ej Eo H.
case X : (p == q) H Eo C'; last first. 
- move=>->{s'} _ /coh_zigzag C'; exists s; split=>//. 
  by exists q, e, h; rewrite !getE.
case=>ls [Ej' Es'] Eo C'. 
set s'' := (tb ->>[self s \+ fresh (self s \+ getso t \+ other s) 
                   \\-> (e :: ls, ls), joint s', other s]).
have E : s' = s'' \< t.
- rewrite (zigSOE C) {1}(cohE C') Es' Eo.
  by rewrite -joinA joinCA /s'' !fsE zigE zipPt. 
rewrite {}E in C' *; exists s''; split=>//; exists q, e, h.
rewrite !getzC //= !getE joinA. 
by split=>//; [exists ls | split; apply/coh_zigzag].
Qed.

Definition try_pop_pre_act := 
  Action trypop_safe_coh trypop_safe_zagzig trypop_step_safe trypop_step_guar 
         trypop_step_frame trypop_step_total.

Lemma trypop_step_erase : int_step_erase (try_pop_pre_act).
Proof.
move=>s1 h1 t1 v1 s2 h2 t2 v2 V E H1 H2; move: H1 H2 V E. 
case=>q1 [e1][hg1][_ J1 _ -> H1 [C1 C1']][q2][e2][hg2][_ J2 _ -> H2 [C2 C2']].
rewrite (cohE C1) (cohE C2) !tb_flatsE J1 J2 -!joinA => V E.
case/(hcancelV V): E=>? {V} V /(hcancelV V) [[?] _ E]; subst q2 e2.
case: (p == q1) H1 H2=>[[ls1][J1' _][ls2][J2' _]|->->]. 
- by rewrite (cohE C1') (cohE C2') !tb_flatsE J1' J2' -!joinA E.
by rewrite (cohE C1) (cohE C2) !tb_flatsE J1 J2 -!joinA E.
Qed.

End TryPopAct. 

Module Exports.
Definition try_pop_act T p p' := IntAction (@trypop_step_erase T p p'). 
End Exports. End TryPopAct.
Export TryPopAct.Exports.


Module ReadSentAct.
Section ReadSentAct.
Variable T : treiberT.
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation coh := (tb_coh T).
Notation U := (tbU T).
Notation getso s := (getso (tbg T) s).
Notation self s := (self (tbg T) s).
Notation other s := (other (tbg T) s).
Notation joint s := (joint (tbg T) s).
Notation guar := (Mod.internal (Treiber T)). 

Definition readsent_step s s' (v : ptr) := exists h : heap, [/\ s' = s, joint s = sent :-> v \+ h & s \In coh].

Lemma readsent_safe_coh : safe_coh coh coh. Proof. by []. Qed.

Lemma readsent_step_safe : step_safe coh readsent_step. Proof. by move=>s s' v [h][]. Qed.

Lemma readsent_safe_zagzig : safe_zagzig coh. Proof. by move=>s s' /coh_zigzag. Qed.

Lemma readsent_step_guar : step_guar guar readsent_step. 
Proof. by move=>s s' v [h][->] _; apply: refl_guar_refl. Qed.

Lemma readsent_step_total : step_total coh readsent_step.
Proof. 
move=>s C; case: (C)=>p [h][g][ls][_ J _ _ _ _].
by exists s, p, (h \+ g); rewrite joinA. 
Qed.

Lemma readsent_step_frame : step_frame coh readsent_step.
Proof. 
move=>s _ t v C [h][-> J _]; exists s; split=>//.
by exists h; rewrite !getE in J *.
Qed.

Definition read_sent_preact := 
  Action readsent_safe_coh readsent_safe_zagzig readsent_step_safe 
         readsent_step_guar readsent_step_frame readsent_step_total.

Lemma readsent_step_erase : int_step_erase read_sent_preact.
Proof.
move=>s1 h1 t1 v1 s2 h2 t2 v2 V E H1 H2.
case: H1 H2 V E=>h [->{t1} J1 C1][h'][->{t2} J2 C2] V E; split=>//.
rewrite (cohE C1) (cohE C2) J1 J2 !tb_flatsE -!joinA in V E. by case/(hcancelV V): E.
Qed.

End ReadSentAct.

Module Exports.
Definition read_sent_act T := IntAction (@readsent_step_erase T). 
End Exports.

End ReadSentAct.

Export ReadSentAct.Exports.

Module ReadNodeAct.
Section ReadNodeAct.
Variables (T : treiberT) (p : ptr).
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation coh := (tb_coh T).
Notation U := (tbU T).
Notation getso s := (getso (tbg T) s).
Notation self s := (self (tbg T) s).
Notation other s := (other (tbg T) s).
Notation joint s := (joint (tbg T) s).
Notation guar := (Mod.internal (Treiber T)). 

Definition readnode_safe s := exists (q p' : ptr) (e : A) (h : heap),
    joint s = sent :-> q \+ p :-> (e, p') \+ h /\ s \In coh.

Definition readnode_step s s' (v : A * ptr) := exists (q p' : ptr) (e : A) (h : heap), 
    [/\ s' = s, v = (e, p'), joint s = sent :-> q \+ p :-> v \+ h & s \In coh].

Lemma readnode_safe_coh : safe_coh coh readnode_safe. Proof. by move=>s [q][p'][e][h][]. Qed.

Lemma readnode_step_safe : step_safe readnode_safe readnode_step.
Proof. by move=>s s' v [q][p'][e][h][_ -> J C]; exists q, p', e, h. Qed. 

Lemma readnode_safe_zagzig : safe_zagzig readnode_safe.
Proof. 
move=> s t [q][p'][e][h][J /coh_zigzag C]. by exists q, p', e, h; rewrite !getE in J *.
Qed.

Lemma readnode_step_guar : step_guar guar readnode_step.
Proof. by move=>s s' v [q][p'][e][h][->-> _]; apply: refl_guar_refl. Qed.

Lemma readnode_step_total : step_total readnode_safe readnode_step.
Proof. by move=>s [q][p'][e][h][J C]; exists s, (e, p'), q, p', e, h. Qed.

Lemma readnode_step_frame : step_frame readnode_safe readnode_step.
Proof. 
move=>s _ t _ _ [q][p'][e][h][->-> J /coh_zigzag C].
by exists s; split=>//; exists q, p', e, h; rewrite !getE in J *.  
Qed.

Definition readnode_preact := 
  Action readnode_safe_coh readnode_safe_zagzig readnode_step_safe 
         readnode_step_guar readnode_step_frame readnode_step_total.

Lemma readnode_step_erase : int_step_erase readnode_preact.
Proof.
move=>s1 h1 t1 v1 s2 h2 t2 v2 V E H1 H2; move: H1 H2 V E.
case=>q1 [p1][e1][x1][->-> J1 C1][q2][p2][e2][x2][->-> J2 C2]; split=>//.
rewrite (cohE C1) (cohE C2) !tb_flatsE J1 J2 -!joinA in V E.
by case/(hcancelV V): E=>_ V' /(hcancelV V') [].
Qed.

End ReadNodeAct.

Module Exports.
Definition read_node_act T p := IntAction (@readnode_step_erase T p). 
End Exports.

End ReadNodeAct.

Export ReadNodeAct.Exports.


Module TryPushAct.
Section TryPushAct.
Variables (pv : nat) (T : treiberT) (p p' : ptr).
Notation A := (tb_set T). 
Variable (e : A).
Notation sent := (tb_sent T). Notation tb := (tb_label T).
Notation st := [encoded_set of seq A]. Notation W := (entangle (Priv pv) (Treiber T)). 
Notation pt_coh := (Mod.coh W). Notation pv_coh := (Mod.coh (Priv pv)).
Notation tb_coh := (Mod.coh (Treiber T)). Notation guar := (Mod.internal W). 
Notation U := (tbU T). Notation tb_getso s := (getso (tbg T) s).
Notation tb_self s := (self (tbg T) s). Notation tb_other s := (other (tbg T) s).
Notation tb_joint s := (joint (tbg T) s). Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s). Notation pv_other s := (other (pvg pv) s).
Notation pv_joint s := (joint (pvg pv) s). 

Lemma pt_cohJ s : s \In pt_coh -> pv_joint s = tt.
Proof.
move=>C; case: C (C)=>_ [s1][s2][->] C1 C2 C.
by rewrite (getC C1 C) pv_cohJ.
Qed.

Definition trypush_safe s := 
  exists h : heap, pv_self s = p :-> (e, p') \+ h /\ s \In pt_coh.

Definition trypush_step s s' v :=
  let: H := tb_self s \+ tb_other s in
  exists (hS : heap) (q : ptr) h, 
    [/\ pv_self s = p :-> (e, p') \+ hS, 
        tb_joint s = sent :-> q \+ h, v = (p' == q),
        if v then 
          [/\ pv_self s' = hS,
              tb_joint s' = sent :-> p \+ p :-> (e, p') \+ h,
              exists ls, tb_self s' = tb_self s \+ fresh H \\-> (ls, e::ls),
              pv_other s' = pv_other s &
              tb_other s' = tb_other s]
        else s' = s &
        s \In pt_coh /\ s' \In pt_coh].

Lemma trypush_safe_coh : safe_coh pt_coh trypush_safe.
Proof. by move=>s [hS][]. Qed.

Lemma trypush_step_safe : step_safe trypush_safe trypush_step.
Proof. by move=>s s' v [hS][q][h][S J _ _][C _]; exists hS. Qed.

Lemma trypush_safe_zagzig : safe_zagzig trypush_safe.
Proof. 
move=>s t [hS][S] /coh_zigzag C; rewrite getE in S.
exists (hS \+ (pv_getso t)); split=>//. by rewrite (getzV (cohSO C)) S joinA.
Qed.

Lemma trypush_step_guar : step_guar guar trypush_step.
Proof. 
move=>s s' _ [h][q][h'][S J -> H][C C'].
case X : (p' == q) H; last by move=>->; apply: refl_guar_refl.
rewrite -{q X}(eqP X) in J *; case=>S' J' [ls S2' O1 O2].
set ag := PushAcqGuar T; set rg := PrivRelGuar pv.
set arg := DoIntGuar (starC _ _) (AcqRelIntGuar ag rg).
right; exists arg; split.
- right; exists (Chan (PrivChanCore pv)), (Chan (TreiberChanCore T)).
  by split; last by [right]; apply/Image_mem. 
split=>//; last by rewrite starC.
rewrite (star_cohE C) (star_cohE C') -!(joinC (tb ->> [_, _, _])).
eexists _, _, _, _; split=>//; exists (p :-> (e, p')); split.
- exists p', p, e, h', ls; rewrite !getE.
  by case: (star_cohP C); case: (star_cohP C').
split; last by case: (star_cohP C). 
by rewrite S S' joinC O1 (pt_cohJ C) (pt_cohJ C'); eauto.
Qed.

Lemma trypush_step_total : step_total trypush_safe trypush_step.
Proof. 
move=>s [h][S] C; case: (star_cohP C)=>E Vh /= C1 C2; move: (validHV Vh)=>V.
case: (C2) (cohVSO C2); rewrite !getE; rewrite E (pt_cohJ C) in Vh V.
move=>q /= [h'][g'][ls] [_ J L K [U M]] _ Vso.
set H : hist _ := tb_self s \+ tb_other s in K U M.
case X : (p' == q); last by exists s, false, h, q, (h' \+ g'); rewrite joinA. 
rewrite -{q X}(eqP X) in L J *.
set s2 := pv ->> [h, tt, pv_other s] \+
          tb ->> [tb_self s \+ fresh H \\-> (ls, e :: ls), sent :-> p \+ p :-> (e, p') \+ h' \+ g', tb_other s].
have V2 : valid s2.
- by rewrite /s2 /valid !fsE /= !fvalidPtUn fvalidPt fdomPt in V *.
have Vh2 : validH s2.
- rewrite /s2 !validHUn !priv_flatsE !tb_flatsE V V2 /= J S in Vh *.
  by rewrite -!joinA -!(joinCA (p :-> _)) -!(joinCA (sent :-> _)) !hvalidPtUn in Vh *. 
exists s2, true, h, p', (h' \+ g'). rewrite eq_refl !joinA. split=>//.
- by rewrite /s2 !getV //=; split=>//; exists ls. 
do !split=>//; eexists _, _; split=>//.
- by split; [move/validHL: Vh2 | exists h, (pv_other s)].
exists p, (p :-> (e, p') \+ h'), g', (e::ls).
rewrite joinA 3!getE -(joinAC _ (tb_other s)); split=>//. 
- by exists p', h'.
- exists ls; rewrite last_fresh // findUnR ?valid_fresh //.
  by rewrite um_domPt inE /= eq_refl um_findPt.
- split; last by apply: fresh_complete.
  by apply: fresh_cont=>//; exists e; right.
by move/validHR: Vh2. 
Qed.

Lemma trypush_step_frame : step_frame trypush_safe trypush_step.
Proof. 
move=>s s' t v [hS][P _][hS'][q][h][E J -> H][C C'].
move/proj1/validL: (star_cohVSO C)=>V. 
rewrite !getE !(getzV (cohSO C)) in E P J H V.
rewrite P -!joinA in E V; case/(hcancelV V): E H C'=>_ _ <-{hS'}.
case: ifP=>[X [P' J' [ls S' O1 O2 C']]|X -> /coh_zigzag]; last first.
- by exists s; split=>//; exists hS, q, h; rewrite !getE.
set K := tb_self s \+ tb_getso t \+ tb_other s. 
set s2 := pv ->> [hS, tt, pv_other s] \+
          tb ->> [tb_self s \+ fresh K \\-> (ls, e :: ls), sent :-> p \+ p :-> (e, p') \+ h, tb_other s].
have V2 : valid s2.
- by case: (star_cohP C)=>-> /validHV; rewrite /s2 /valid !fsE /= !fvalidPtUn.
have E : s' = s2 \< t.
- rewrite /s2 {1}(star_cohE C') (star_zigSOE C) zigUnPt !fsE !zigE !zipPt.
  by rewrite (pt_cohJ C') -(joinC hS) -P' -J' -O1 -O2 joinCA joinA S'. 
rewrite E in C'; exists s2; split=>//; exists hS, q, h. 
rewrite !(getzV (cohSO C')) !(getzV (cohSO C)) !getE // /s2 !getV //=. 
split=>//; last by split; apply/coh_zigzag.
by split=>//; exists ls; rewrite joinA. 
Qed.

Definition try_pop_pre_act := 
  Action trypush_safe_coh trypush_safe_zagzig trypush_step_safe trypush_step_guar 
         trypush_step_frame trypush_step_total.


Lemma trypush_step_erase : int_step_erase try_pop_pre_act.
Proof.
move=>s1 h1 t1 v1 s2 h2 t2 v2 V E H1 H2; move: H1 H2 V E.
case=>hs1 [q1][h1'][S1 J1 -> H1 [C1 C1']][hs2][q2][h2'][S2 J2 -> H2 [C2 C2']].
move: (cohV C1) (cohV C2) (cohV C1') (cohV C2').
rewrite (star_cohE C1) (star_cohE C2) (star_cohE C1') (star_cohE C2').
rewrite (pt_cohJ C1) (pt_cohJ C2) (pt_cohJ C1') (pt_cohJ C2') => V1 V2 V1' V2'.
rewrite !flatsUn // !priv_flatsE !tb_flatsE // S1 S2 J1 J2.
rewrite -!joinA -!(joinCA (sent :-> _)) => V E.
case/(hcancelV V): E J1 J2 H1 H2 => <-{q2} _ E J1 J2.  
case: (p' == q1)=>[[->->][ls1 _ -> _][->->][ls2 _ -> _]|->->].
- by rewrite -!joinA -!(joinCA (p :-> _)) -!(joinCA (sent :-> _)) E.
by rewrite S1 S2 J1 J2 -!joinA -!(joinCA (sent :-> _)) E. 
Qed.

End TryPushAct.

Module Exports.
Definition try_push_act hp T p p' e := IntAction (@trypush_step_erase hp T p p' e). 
End Exports.

End TryPushAct.

Export TryPushAct.Exports.


(***********************************)
(* Stability properties of Treiber *)
(***********************************)

Section Stability.
Variable T : treiberT.
Notation A := (tb_set T). Notation sent := (tb_sent T).
Notation tb := (tb_label T). Notation st := [encoded_set of seq A].
Notation coh := (tb_coh T). Notation U := (tbU T).
Notation getso s := (getso (tbg T) s). Notation self s := (self (tbg T) s).
Notation other s := (other (tbg T) s). Notation joint s := (joint (tbg T) s).
Notation guar := (Mod.internal (Treiber T)).  Notation W := (Treiber T).

(* monotonicity of histories wrt to rely *)
Lemma treiber_mrely s1 s2 : mrely W s1 s2 -> [other s1 <<= other s2].
Proof.
move=>H; case: (mrely_coh H)=>C1 C2; split.
- by rewrite (validR (cohVSO C1)) (validR (cohVSO C2)).
case: H.
- case=>[|[g]][]; first by move/transp_inj=>->; exists Unit; rewrite unitR. 
  move=><- [p][p'][e][h][ls][J1 J2 -> E].
  by rewrite !(loc_transp, env_transp); eexists _.
case=>ch [x][/mem_chan][_][-> <-] /=. 
case; last by case=>_ /transp_inj<- _; exists Unit; rewrite unitR.
case=>p [p'][e][h][ls][_ _ _]. by rewrite !loc_transp !env_transp => -><-; eexists _.
Qed.

Lemma hist_other s1 s2 : menv_steps W s1 s2 -> [other s1 <<= other s2].
Proof.
case=>n; elim: n s1 s2 => [|n IH] s1 s2 /=; first by case=>->. 
by case=>k [] /treiber_mrely H /IH; apply: hist_trans H.
Qed.

(* stability of read *)
Lemma readnode_mrely (p p' q1 : ptr) (e : A) (h1 : heap) s1 s2: 
        mrely W s1 s2 ->
        joint s1 = sent :-> q1 \+ p :-> (e, p') \+ h1 ->
        exists (q2 : ptr) (h2 : heap), joint s2 = sent :-> q2 \+ p :-> (e, p') \+ h2.
Proof.
move=>H J; case: (mrely_coh H)=>C1 C2; case: H.
- case=>[[]|[g][<-]]; first by move/transp_inj=><-; exists q1, h1.
  case=>p1 [p1'][e'][h][ls][E J2 _ _ _ _]; move: (tb_cohJ C1) => V.
  rewrite !getE J -!joinA in V E J2; case/(hcancelV V): E J2 => <- _ <- ->.
  by rewrite joinA; eexists _, _.
case=>ch [x][/mem_chan][_][-> <-] /=.
case; last by case=>_ /transp_inj <-; exists q1, h1.
case=>p1 [p1'][e'][h][ls][_ E J2 _ _ _]; move: (tb_cohJ C1) => V.
rewrite !getE J -!joinA in V E J2; case/(hcancelV V): E J2 => <- _ <- ->.
by rewrite joinA joinAC joinA -joinA; eexists _, _.
Qed.

Lemma readnode_steps (p p' q1 : ptr) (e : A) h1 s1 s2 : 
        menv_steps W s1 s2 ->
        joint s1 = sent :-> q1 \+ p :-> (e, p') \+ h1 ->
        exists (q2 : ptr) (h2 : heap), joint s2 = sent :-> q2 \+ p :-> (e, p') \+ h2.
Proof.
case=>n; elim: n q1 h1 s1 s2 => [q1 h1|n IH q1 h1] s1 s2 /=.
- by case=>-> _; exists q1, h1. 
by case=>t [/readnode_mrely] H M /H [q2][h2]; apply: IH M.
Qed.

(* stack with null sentinel is empty *)
Lemma null_sent s h : let: H := self s \+ other s in
        s \In Mod.coh W -> joint s = sent :-> null \+ h -> [[::] in H at last_key H].
Proof.
move=>C J; case: (C); rewrite {}J=>[p][h'][g][ls][Es] E L K [S Ct] V. 
rewrite /validH {1}Es {Es} tb_flatsE -joinA in V E L K S Ct *.
case/(hcancelV V): E L K S Ct=><-{p} {V} V E. by move: E V => -> /validL V /(llist_null V) [->].
Qed.

End Stability.


(***********************)
(* Treiber pop program *)
(***********************)

Section TreiberPop. 
Variable T : treiberT.
Notation A := (tb_set T). Notation sent := (tb_sent T). 
Notation tb := (tb_label T). Notation st := [encoded_set of seq A].
Notation coh := (tb_coh T). Notation U := (tbU T).
Notation getso s := (getso (tbg T) s). Notation self s := (self (tbg T) s).
Notation other s := (other (tbg T) s). Notation joint s := (joint (tbg T) s).
Notation guar := (Mod.internal (Treiber T)). Notation W := (Treiber T).

(* first a number of stabilized actions *)

Program Definition read_sentinel : 
  {h}, STsep [W] (fun i => self i = Unit /\ [h <<= other i],
                  fun (p : ptr) m => exists t, 
                    [/\ last_key h <= t,
                        self m = Unit, [h <<= other m] &
                        if p == null then [[::] in other m at t] 
                        else exists (q p' : ptr) e ls h' ,  
                            [e :: ls in other m at t] /\ joint m = sent :-> q \+ p :-> (e, p') \+ h']) :=
  Do (act (read_sent_act T)).
Next Obligation. 
apply: gh=>i1 h [S H] C; apply: val_act=>i2 M.
rewrite (menvs_loc M) in S; move: (hist_trans H (hist_other M))=>{H} H.
split=>[|p _ i3 [h1][-> J {C} C] {M} M]; first by case/menvs_coh: M. 
exists (last_key (other i2)).
move: (hist_other M)=>O; rewrite -(menvs_loc M) (histK H).
split=>//; first by apply: hist_trans H O. 
case X : (p == null). 
- rewrite (eqP X) in J; move/(null_sent C): J.
  by rewrite S unitL; apply: histE O.
case: C=>q [h'][g][ls][E' E L G _] V.
rewrite J -joinA in E; rewrite /validH {}E' tb_flatsE J in V.
case/(hcancelV V): E L J G => <-{q V} _ ->{h1}.
case/(first_exists (negbT X)) => e [p'][h''][->->{h'}] _. 
rewrite -!joinA joinA S unitL; case/(readnode_steps M)=>q [h' J]. 
by exists q, p', e, (behead ls), h'; split=>//; apply: histE O G.
Qed.

Program Definition read_node p : 
  {h p' e}, STsep [W] (fun i => exists (q1 : ptr) h1, 
                         [/\ self i = Unit, [h <<= other i], p != null & 
                             joint i = sent :-> q1 \+ p :-> (e, p') \+ h1],
                       fun v m => exists (q1 : ptr) h1, 
                         [/\ self m = Unit, [h <<= other m], v = (e, p') & 
                             joint m = sent :-> q1 \+ p :-> (e, p') \+ h1]) :=
  Do (act (read_node_act T p)).
Next Obligation.
apply: gh=>i1 [h][p' e][q1][h1][/= S H N J] _; apply: val_act=>i2 M.
move: (hist_trans H (hist_other M))=>{H} H; rewrite (menvs_loc M) in S. 
case/(readnode_steps M): J=>{q1 h1} q1 [h1 J]. 
split=>[|_ _ i3 [q2][p2][e2][h2][->-> J2] C {M} M].
- by exists q1, p', e, h1; split=>//; case/menvs_coh: M.
move: (tb_cohJ C)=>V; rewrite J -!joinA -(menvs_loc M) in V J2 *.
case/(hcancelV V): J2=>_ {V} V /(hcancelV V) [<- _ _] {V}. 
case/(readnode_steps M): J=>{q1 h1} q1 [h1] J.
by exists q1, h1; split=>//; apply: hist_trans H (hist_other M).
Qed.

(* A vanilla, non-framed specification.   *)
(* For framed version, see stack_framed.v *)
Program Definition try_pop p p' : 
  {h e}, STsep [W] (fun i => exists (q1 : ptr) h1, 
                      [/\ self i = Unit, [h <<= other i], p != null & 
                          joint i = sent :-> q1 \+ p :-> (e, p') \+ h1],
                    fun v m => exists (q1 : ptr) h1, 
                      [/\ [h <<= other m], joint m = sent :-> q1 \+ p :-> (e, p') \+ h1 &
                          if v then 
                            exists t ls, last_key h < t /\ self m = t \\-> (e :: ls, ls)
                          else self m = Unit]) :=
  Do (act (try_pop_act T p p')).
Next Obligation.
apply: gh=>i1 [h e][q1][h1][/= S H N J] _; apply: val_act=>i2 M.
move: (hist_trans H (hist_other M))=>{H} H; rewrite (menvs_loc M) in S.
case/(readnode_steps M): J=>{q1 h1} q1 [h1] J.
split=>[|v i3 i4 [q2][e2][h2][_ J2 O ->{v} H'][C _]].
- by exists q1, e, h1; split=>//; case/menvs_coh: M.
move: (tb_cohJ C)=>V; rewrite J -!joinA in V J2.
case/(hcancelV V): J2 H' O=><- {V} V /(hcancelV V) [[<-]] _ <-{V q2 e2 h2}.
case: (p == q1); last first.
- move=>->{i3} _ {M} M; case/(readnode_steps M): J=>{q1 h1} q1 [h1 J].
  exists q1, h1; rewrite -(menvs_loc M); split=>//.
  by apply: hist_trans H (hist_other M).
case=>ls [{J} J G O] {M} M; rewrite {}S !unitL in G.
rewrite -{}O in H G; rewrite -(menvs_loc M).
case/(readnode_steps M): J=>{q1 h1} q1 [h1 J].
exists q1, h1; split=>//; first by apply: hist_trans H (hist_other M).
by exists (fresh (other i3)), ls; split=>//; apply: histK H.
Qed.

(* now the main pop program *)

Definition pop_tp := 
  {h}, STsep [W] (fun i => self i = Unit /\ [h <<= other i],
                  fun v m => 
                    [/\ [h <<= other m] &
                        if v is Some e then 
                          exists t ls, last_key h < t /\ 
                                       self m = t \\-> (e :: ls, ls)
                        else exists t, [/\ last_key h <= t, self m = Unit & 
                                           [[::] in other m at t]]]).

Program Definition pop :=
  ffix (fun (loop : unit -> pop_tp) xx =>
    Do (p <-- read_sentinel; 
        if p == null then ret None
        else 
          res <-- read_node p;
          ok <-- try_pop p res.2;
          if ok then ret (Some res.1) else loop tt)) tt.
Next Obligation.
apply: gh=>i1 h [S H] C; apply: step; apply: (gh_ex h).
apply: val_do=>// p i2 [t2][K {S H} S H]. 
case X : {1 2}(p == null)=>/= {i1 C}. 
- move=>He C; apply: val_ret=>i3 M.
  rewrite -(menvs_loc M); move: (hist_other M)=>O.
  by split=>//; [apply: hist_trans H O | exists t2; split=>//; apply: histE O He].
case=>{K} q [p'][e][ls][h'][_ J] C.
apply: step; apply: (gh_ex h); apply: (gh_ex p'); apply: (gh_ex e). 
apply: val_do; first by rewrite X; exists q, h'. 
move=>_ i3 {q ls S H C J} [q][h1][S H] -> J C /=.  
apply: step; apply: (gh_ex h); apply: (gh_ex e).
apply: val_do; first by rewrite X; exists q, h1. 
case=>i4 {q h1 H J S C} [q][h1][H J]; last first. 
- by move=>S C; apply: (gh_ex h); apply: val_do.
case=>t [ls][K] S C; apply: val_ret=>i5 M; split=>//.
- by apply: hist_trans H (hist_other M).
by exists t; rewrite -(menvs_loc M); exists ls. 
Qed.

End TreiberPop.

(************************)
(* Treiber Push program *)
(************************)

Section TreiberPush.
Variables (pv: nat) (T : treiberT).
Notation A := (tb_set T). Notation sent := (tb_sent T).
Notation l := (tb_label T). Notation st := [encoded_set of seq A].
Notation W := (entangle (Priv pv) (Treiber T)). 
Notation pt_coh := (Mod.coh W). Notation pv_coh := (Mod.coh (Priv pv)).
Notation tb_coh := (Mod.coh (Treiber T)). Notation guar := (Mod.internal W). 
Notation U := (tbU T). Notation tb_getso s := (getso (tbg T) s).
Notation tb_self s := (self (tbg T) s). Notation tb_other s := (other (tbg T) s).
Notation tb_joint s := (joint (tbg T) s). Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s). Notation pv_other s := (other (pvg pv) s).

Lemma phist_other s1 s2 : 
        menv_steps W s1 s2 -> [tb_other s1 <<= tb_other s2].
Proof.
move=>M; case: (menvs_coh M)=>C1 C2.
case: {C1 C2} (star_cohP C1) (star_cohP C2) M => E1 _ P1 _ [E2 _ P2 _].
rewrite {1}E1 {1}E2.
move/(menvs_split (entangle_extL _ _) P1 P2)/proj2/hist_other.
by rewrite !getE. 
Qed.

(* stabilized action try_push action *)

Program Definition try_push (p p' : ptr) (e : A) : 
  {h hS}, STsep [W] (fun i => 
                       [/\ pv_self i = p :-> (e, p') \+ hS, 
                           tb_self i = Unit & [h <<= tb_other i]],
                     fun v m =>  
                       if v then exists t ls, 
                         [/\ pv_self m = hS, last_key h < t, 
                             tb_self m = t \\-> (ls, e :: ls) & 
                             [h <<= tb_other m]]
                       else 
                         [/\ pv_self m = p :-> (e, p') \+ hS, 
                             tb_self m = Unit & [h <<= tb_other m]]) :=
  Do (act (try_push_act pv p p' e )).
Next Obligation.
apply: gh=>i [h hS][/= P S H] _; apply: val_act=>i2 M; case: (menvs_coh M)=>_ C.
rewrite (menvs_loc M) in P S; move: (hist_trans H (phist_other M))=>{i M H} H.
split=>[|_ m1 m2 [hS'][q][h'][E J] ->]; first by exists hS.
(* the next two lines would disappear if actions used binarify *)
move: (validL (proj1 (star_cohVSO C)))=>V; rewrite P in E V.
case/(hcancelV V): E J=>_ _ <-{V hS'}.
case: ifP; last first. 
- move=>_ J ->{m1} _ M; rewrite -(menvs_loc M). 
  by split=>//; apply: hist_trans (phist_other M). 
move/eqP=><-{q} J [P' J'][ls] T' _ O _ M. 
exists (fresh (tb_self i2 \+ tb_other i2)), ls.
rewrite (menvs_loc M) S !unitL in P' J' T' *; split=>//.
- by apply: histK H. 
by apply: hist_trans (phist_other M); rewrite O. 
Qed.

(* Introducing the allocator *)
Variables (al : nat) (asent alk : ptr).

(* Use a concrete, CAS-based allocator *)
Notation Alloc := (AllocCAS al asent alk). Notation V := (entangle W Alloc). 
Notation pta_coh := (Mod.coh V). Notation al_coh := (al_coh al asent alk). 

(* now the main push program *)
Let push_loop_tp (e : A) (p : ptr) := 
  {h hS}, STsep [V] (fun i => exists B (v : B),
                      [/\ pv_self i = p :-> v \+ hS, tb_self i = Unit &
                          [h <<= tb_other i]], 
                    fun (_ : unit) m => exists t ls,
                      [/\ pv_self m = hS, tb_self m = t \\-> (ls, e :: ls), 
                          [h <<= tb_other m] & last_key h < t]).

Program Definition push e :   
  {h hS}, STsep [V] (fun i => [/\ tb_self i = Unit, pv_self i = hS & [h <<= tb_other i]],
                     fun (_ : unit) m => exists t ls, 
                       [/\ tb_self m = t \\-> (ls, e :: ls), pv_self m = hS,
                           [h <<= tb_other m] & last_key h < t]) := 
  Do (p <-- [alloc pv al asent alk];
      ffix (fun (loop : unit -> push_loop_tp e p) xx =>
        Do (p' <-- [read_sentinel T]; 
            [priv_write pv p (e, p')];;
            ok <-- [try_push p p' e];
            if ok then ret tt else loop tt)) tt).
Next Obligation. exact: injLE (entangleAC _ _ _). Defined.
Next Obligation. exact: injRE (entangleAC _ _ _). Defined.
Next Obligation. exact: injLE (entangle_apartA _ _ _). Defined.
Next Obligation. exact: injLE (erefl _). Defined.
Next Obligation.
apply: gh=>i [h hS][B][v] X C.
case: C (C) X=>_ [_][xa][->{i}][_][xp][xt][->] Cp Ct Ca C [P S H].
rewrite (getC Cp C) !(getC Ct C) /= in P S H. 
rewrite -!joinA joinCA in C *. 
apply: step; apply: val_extend=>//; apply: (gh_ex h); apply: val_do=>//.
move=>p' {xt S H C Ct Ca} xt [t][K] S H _ Ct s C M. 
case: {M} (menvs_coh M) (M) C=>_ [_][xp1][xa1][->{s}] Cp1 Ca M C.
case/(menvs_split (injLE _ (erefl _)) Cp Cp1): M=>/= M _.
have {M P} P : pv_self xp1 = p :-> v \+ hS by rewrite -(menvs_loc M).
rewrite -joinCA in C *. apply: step; apply: val_extend=>//.
apply: (gh_ex hS); apply: val_do; first by exists B, v. 
case=>{xp xp1 xa C Cp Cp1 Ca P} xp P Cp s /= C M.
case: {M} (menvs_coh M) (M) C=>_ [_][xt1][xa][->{s}] Ct1 Ca M C.
case/(menvs_split (injLA _ (erefl _)) Ct Ct1): M=>/= {xa1} M _.
have {S} S : tb_self xt1 = Unit by rewrite -(menvs_loc M).
have {xt Ct M H} H : [h <<= tb_other xt1] by apply: hist_trans H (hist_other M).
rewrite joinA in C *.
apply: step; apply: val_extend; first by apply/(star_coh_prec C).
apply: (gh_ex h); apply: (gh_ex hS); apply: val_do.
- by move=>C'; rewrite (getC Cp C') !(getC Ct1 C'). 
move=>b s X Y; case: Y (Y) X=>_ [xp1][xt][->{s}] {Cp} Cp1 Ct Y X.
move=>xa1 /= {C} C M; case: (menvs_coh M)=>_ {M xa Ca} Ca1.
rewrite (getC Cp1 Y) !(getC Ct Y) in X. 
case: b X; last first.
- case=>{P S H} P S H. 
  apply: (gh_ex h); apply: (gh_ex hS); apply: val_do=>//.
  move=>{C} C; exists (prod A ptr), (e, p').
  by rewrite (getC Cp1 C) !(getC Ct C).
case=>{xp xt1 Ct1 P K S H} t' [ls][P K S H]. 
apply: val_ret=>//= m M; rewrite -(menvs_loc M).
rewrite (getC Cp1 C) (getC Ct C).
exists t', ls; split=>//.
(* introduce a coh clause in val_ret, to remove the need for *)
(* explicitly getting coherence out of menvs_coh *)
case: {M} (menvs_coh M) (M)=>_ /= X'.
case: X' (X')=>_ [s][xa][->{m}] Y'; case: Y' (Y')=>_ [xp][xt1][->].
move=>Cp Ct1 Y' {Ca1} Ca {C} C. 
case/(menvs_split (injLE _ (erefl _)) Y Y').
case/(menvs_split (injLE _ (erefl _)) Cp1 Cp)=> _ M _.
by rewrite (getC Ct1 C); apply: hist_trans H (hist_other M).
Qed.

(* requires more cleanup *)
Next Obligation.
apply: gh=>i [h hS] X C.
case: C (C) X =>_ [_][xa][->][_][xp][xt][->] Cp Ct Ca C /= [P Ps H].
rewrite !(getC Ct C) !(getC Cp C) /= in P Ps H.
rewrite joinAC /V /= starAC in C *.
apply: step; apply: val_extend; first by apply/(star_coh_prec C). 
apply: (gh_ex (pv_self xp)); apply: val_do. 
- by move=>Cpa; rewrite (getC Cp Cpa).
move=>x m [B][v] Y X {Cp Ca C}.
case: X (X) Y=>_ [xp1][xa1][->{m}] /= Cp Ca X S xt1 C M. rewrite {X} (getC Cp X) in S.
(* introduce a coh clause in val_extend to remove the need for *)
(* explicitly getting coherence out of menvs_coh *)
case: (menvs_coh M)=>_ /= {Ct} Ct.
have {P} P : tb_self xt1 = Unit by rewrite -(menvs_loc M).
have {H M} H : [h <<= tb_other xt1] by apply: hist_trans H (hist_other M).
apply: (gh_ex h); apply: (gh_ex (pv_self xp)).
apply: val_do=>[_|]; first by exists B, v; rewrite (getC Cp C) !(getC Ct C).
case=>m [t][ls][P2 S2 H2 K2]; exists t, ls; rewrite P2; split=>//.
Qed.

End TreiberPush.



(*******************************************************************)
(* Instances of decorate and mod_unit structures for Treiber stack *)
(*******************************************************************)

Module TreiberDecorate.
Section TreiberDecorate.
Variables (T : treiberT).
Notation A := (tb_set T). Notation sent := (tb_sent T).
Notation tb := (tb_label T). Notation coh := (tb_coh T).
Notation U := (tbU T). Notation getso s := (getso (tbg T) s).
Notation self s := (self (tbg T) s). Notation other s := (other (tbg T) s).
Notation joint s := (joint (tbg T) s). Notation guar := (Mod.internal (Treiber T)). 

Definition no_garbage (g : U * heap) := (forall t s1 s2, find t g.1 = Some (s1, s2) -> 
  [/\ t = 0 & s1 = s2] \/ exists v, s2 = v::s1) -> g.2 = Unit.

Definition treiber_dec (g : U * heap) s := exists h (ls0 ls: seq A) (p: ptr),
    [/\ s = tb ->> [(0 \\-> (ls0, ls0) : U) \+ g.1, sent :-> p \+ h \+ g.2, Unit] & llist p ls h] /\ s \In coh.

Lemma hist_init0 p ls1 ls2 h g : llist p ls1 h -> 
        tb ->> [0 \\-> (ls2, ls2) : U, sent :-> p \+ h \+ g, Unit] \In coh -> ls1 = ls2. 
Proof.
move=>L C; move: (cohH C) (cohV C)=>Vh V; rewrite /validH tb_flatsE in Vh.
case: C=>x' [r'][g'][ls']; rewrite !getV //=; case=>_ E L' K' _ _. 
rewrite -!joinA in E Vh; case/(hcancelV Vh): E=>? {Vh} Vh E; subst x'.
case: (llist_prec Vh E L L')=>? _ _; subst ls'. 
by case: K'=>x; rewrite unitR last_keyPt um_findPt; case=>_ ->.  
Qed.

Lemma hist_emp (g : U) : ~~ empb g -> valid g -> exists t v, find t g = Some v.
Proof.
elim/um_indf: g=>[_||k v f _ V _ _ _] /=. 
- by rewrite -[valid _]negbK; move/(elimN idP); rewrite invalidE. 
- by case: empbP.
by exists k, v; rewrite um_findPtUn.
Qed.

Lemma hist_init1 g ls (h : heap) : 
        ~~ empb g -> tb ->> [(0 \\-> (ls, ls) : U) \+ g, h, Unit] \In coh -> 
        exists ls', find 1 g = Some (ls, ls').
Proof.
move=>N C; move: (cohV C) (cohH C)=>V Vh; rewrite /validH tb_flatsE in Vh.
have Vn : valid (0 \\-> (ls, ls) \+ g).
- by move: (cohVSO C); rewrite getV //=; move/validL. 
have Vg : valid g by move/validR: Vn.  
case: C=>x' [r'][g'][ls'][]; rewrite !getV //= =>_ E L' K' [_ C'] _.
rewrite unitR in K' C'; case: C'; case =>_ /(_ 0) Wf1 /(_ 1) Wf2.
rewrite !um_domPtUn !inE !Vn /= um_findPtUn // in Wf1 Wf2 *.
rewrite (_ : last_key (0 \\-> (ls, ls) \+ g) = last_key g) in Wf2; last first.
- apply: max_key_last.  
  - rewrite um_domPtUn inE Vn /= orbC; case X: (_ \in _)=>//=.
    by move: (last_key_dom Vg (negbT X)) N=>->; case: empbP.
  move=>y; rewrite um_domPtUn inE Vn /=; case: eqP=>[<-//|_].
  by apply: last_key_max.
suff: 0 < last_key g.
- move/Wf2=>D; case: (Wf1 (erefl _) D)=>s1 [s][s2][][_ <-].
  by rewrite findUnR // D => ->; exists s2.
case: (hist_emp N Vg)=>/= t [v] /find_some D. 
move/last_key_max: (D); apply: leq_trans. case: t D=>//; case: validUn Vn=>// _ _ /(_ 0).
by rewrite um_domPt inE; move/(_ (erefl _))/negbTE=>->. 
Qed.

Lemma dec_prec g h1 h2 s1 s2 : valid (flats s1 \+ h1) -> flats s1 \+ h1 = flats s2 \+ h2 ->
        treiber_dec g s1 -> treiber_dec g s2 -> s1 = s2.
Proof. 
move=>V Ef [h][ls0][ls][p][[E1] L1] C1 [h'][ls0'][ls'][p'][[E2] L2]C2.
subst s1 s2; rewrite !tb_flatsE in V C1 C2 Ef; rewrite -!joinA in V Ef.
case/(hcancelV V): Ef=>? {V} V Ef; subst p'. rewrite -(joinC h1) -(joinC h2) in Ef V.
case: (llist_prec V Ef L1 L2)=>?? {Ef} Ef; subst ls' h'.
case E: (empb g.1).
- case: empbP E=>// E _; rewrite E !unitR in C1 C2 *.
  by rewrite -(hist_init0 L1 C1) -(hist_init0 L2 C2).
case: (hist_init1 (negbT E) C1)=>ls1 F1.
case: (hist_init1 (negbT E) C2)=>ls2 F2.
by rewrite F1 in F2; case: F2=><-. 
Qed.

Lemma dec_inj g1 g2 s : 
        treiber_dec g1 s -> treiber_dec g2 s -> g1 = g2.
Proof. 
move=>[h][ls0][ls][p][[E1] L1]C1 [h'][ls0'][ls'][p'][[E2] L2]_; rewrite E1 in E2.
move/cohH: (C1); rewrite E1 /validH tb_flatsE=> V.
move/cohSO: (C1); rewrite E1 /validSO !fsE zipPt allvalidPt unitR=>V1. 
rewrite ![_ \+ _ \+ g1.2]joinAC ![_ \+ _ \+ g2.2]joinAC in V E1 E2.
case/scancelPt: E2=>G1; rewrite -!joinA in V *.
case/(hcancelV V)=>Z2 V2 E2; subst p'. rewrite -(joinC h) -(joinC h') in V2 E2.
case:(llist_prec V2 E2 L1 L2)=>_ _ G2 {E1 V V2 E2}.
suff G1': g1.1 = g2.1 by move: (g1) (g2) G1' G2=>[x1 y1][x2 y2]/=->->. 
by case: (gen_cancel V1 G1).
Qed.

Lemma dec_coh g s : treiber_dec g s -> s \In Mod.coh (Treiber T).
Proof. by case=>?[?][?][?][_]. Qed.

Lemma dec_env g1 s1 :
        treiber_dec g1 s1 -> env s1 = sig (Mod.coh (Treiber T)).
Proof. by move=>[h][ls0][ls][p][[E1] L1]C1; by rewrite E1 fsE. Qed.

Lemma dec_onto s2 : 
        s2 \In coh -> env s2 = sig (Mod.coh (Treiber T)) -> 
        exists g2, treiber_dec g2 s2.
Proof. 
move=>C E; case: (C)=>/=p[h][g][ls][E2] Ej L H1 [H2 H3] Vh.
suff G0 : other s2 = Unit; last by rewrite -[s2]st_eta E /= getsoPt.
rewrite G0 unitR in H1 H2 H3.
case: (H3)=>Wf K; case: Wf=>[[ls']G1] G2.
suff G4: exists H', self s2 = 0 \\-> (ls', ls') \+ H'.
- case: G4=>H Eh; exists (H, g), h, ls', ls, p=>/=; do![split=>//].
  by rewrite -Eh -Ej -G0.
case/find_some/um_eta: (G1)=>v [X] ->; exists (free 0 (self s2)).
by rewrite G1 in X; case: X=><-. 
Qed.

End TreiberDecorate.

Module Exports.
Notation treiber_dec := treiber_dec.

Section Exports.
Variable T : treiberT.
Notation A := (tb_set T). Notation sent := (tb_sent T).
Notation tb := (tb_label T). Notation coh := (tb_coh T). Notation U := (tbU T).

Definition TreiberDecorateMixin := DecorateMixin (@dec_prec T) (@dec_inj T) (@dec_coh T) (@dec_onto T) (@dec_env T).
Canonical TreiberDecorate := Eval hnf in Decorate (@treiber_dec T) TreiberDecorateMixin.

End Exports. End Exports.
End TreiberDecorate.

Export TreiberDecorate.Exports.

